/*
 * Copyright 2017 Google Inc. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.android.installreferrer.api;

import android.content.Context;
import android.os.RemoteException;
import android.support.annotation.IntDef;
import android.support.annotation.NonNull;
import android.support.annotation.UiThread;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;

/** Main interface for communication between the library and user application code. */
public abstract class InstallReferrerClient {

  /** Possible return codes. */
  @Retention(RetentionPolicy.SOURCE)
  @IntDef({
    InstallReferrerResponse.SERVICE_DISCONNECTED,
    InstallReferrerResponse.OK,
    InstallReferrerResponse.SERVICE_UNAVAILABLE,
    InstallReferrerResponse.FEATURE_NOT_SUPPORTED,
    InstallReferrerResponse.DEVELOPER_ERROR
  })
  public @interface InstallReferrerResponse {
    /**
     * Play Store service is not connected now - potentially transient state.
     *
     * <p>E.g. Play Store could have been updated in the background while your app was still
     * running. So feel free to introduce your retry policy for such use case. It should lead to a
     * call to {@link #startConnection(InstallReferrerStateListener)} right after or in some time
     * after you received this code.
     */
    int SERVICE_DISCONNECTED = -1;
    /** Success. */
    int OK = 0;
    /** Could not initiate connection to the Install Referrer service. */
    int SERVICE_UNAVAILABLE = 1;
    /** Install Referrer API not supported by the installed Play Store app. */
    int FEATURE_NOT_SUPPORTED = 2;
    /** General errors caused by incorrect usage */
    int DEVELOPER_ERROR = 3;
  }

  /** Builder to configure and create a InstallReferrer instance. */
  public static final class Builder {
    private final Context mContext;

    private Builder(Context context) {
      mContext = context;
    }

    /**
     * Creates a Install Referrer client instance.
     *
     * <p>After creation, it will not yet be ready to use. You must initiate setup by calling {@link
     * #startConnection(InstallReferrerStateListener)} and wait for setup to complete.
     *
     * @return InstallReferrer instance
     * @throws IllegalArgumentException if Context were not set.
     */
    @UiThread
    public InstallReferrerClient build() {
      if (mContext == null) {
        throw new IllegalArgumentException("Please provide a valid Context.");
      }

      return new InstallReferrerClientImpl(mContext);
    }
  }

  /**
   * Constructs a new {@link Builder} instance.
   *
   * @param context It will be used to get an application context to bind to the Install Referrer
   *     service.
   */
  @UiThread
  public static Builder newBuilder(@NonNull Context context) {
    return new Builder(context);
  }

  /**
   * Checks if the client is currently connected to the service, so that requests to other methods
   * will succeed.
   *
   * <p>Returns true if the client is currently connected to the service, false otherwise.
   *
   * <p>Note: It also means that Install Referrer API is available for the current device.
   */
  @UiThread
  public abstract boolean isReady();

  /**
   * Starts up InstallReferrerClient setup process asynchronously. You will be notified through the
   * {@link InstallReferrerStateListener} listener when the setup process is complete.
   *
   * @param listener The listener to notify when the setup process is complete.
   */
  @UiThread
  public abstract void startConnection(@NonNull final InstallReferrerStateListener listener);

  /**
   * Close the connection and release all held resources such as service connections.
   *
   * <p>Call this method once you are done with this InstallReferrerClient reference.
   */
  @UiThread
  public abstract void endConnection();

  /**
   * Get installation referrer details. This method uses information stored by the Google Play Store
   * app without initiating a network request.
   *
   * @return ReferrerDetails The {@link ReferrerDetails} with all install referrer information.
   */
  @UiThread
  public abstract ReferrerDetails getInstallReferrer() throws RemoteException;
}
