/*
 * Copyright 2017 Google Inc. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.android.installreferrer.api;

import static java.lang.annotation.RetentionPolicy.SOURCE;

import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.ServiceConnection;
import android.content.pm.PackageInfo;
import android.content.pm.PackageManager;
import android.content.pm.ResolveInfo;
import android.os.Bundle;
import android.os.IBinder;
import android.os.RemoteException;
import android.support.annotation.IntDef;
import android.support.annotation.NonNull;
import com.android.installreferrer.commons.InstallReferrerCommons;
import com.google.android.finsky.externalreferrer.IGetInstallReferrerService;
import java.lang.annotation.Retention;
import java.util.List;

class InstallReferrerClientImpl extends InstallReferrerClient {
  private static final String TAG = "InstallReferrerClient";

  // Version 8.3.73 - U_MR_Release
  private static final int PLAY_STORE_MIN_APP_VER = 80837300;

  private static final String SERVICE_PACKAGE_NAME = "com.android.vending";
  private static final String SERVICE_NAME =
      "com.google.android.finsky.externalreferrer.GetInstallReferrerService";
  private static final String SERVICE_ACTION_NAME =
      "com.google.android.finsky.BIND_GET_INSTALL_REFERRER_SERVICE";

  /** Possible client/referrer service relationship states. */
  @IntDef({
    ClientState.DISCONNECTED,
    ClientState.CONNECTING,
    ClientState.CONNECTED,
    ClientState.CLOSED
  })
  @Retention(SOURCE)
  public @interface ClientState {
    /**
     * This client was not yet connected to install referrer service or was already disconnected
     * from it.
     */
    int DISCONNECTED = 0;
    /** This client is currently in process of connecting to install referrer service. */
    int CONNECTING = 1;
    /** This client is currently connected to install referrer service. */
    int CONNECTED = 2;
    /** This client was already closed and shouldn't be used again. */
    int CLOSED = 3;
  }

  private @ClientState int mClientState = ClientState.DISCONNECTED;

  /** Context of the application that initialized this client. */
  private final Context mApplicationContext;

  /** Service binder. */
  private IGetInstallReferrerService mService;

  /** Service connection */
  private ServiceConnection mServiceConnection;

  public InstallReferrerClientImpl(@NonNull Context context) {
    mApplicationContext = context.getApplicationContext();
  }

  @Override
  public boolean isReady() {
    return mClientState == ClientState.CONNECTED && mService != null && mServiceConnection != null;
  }

  @Override
  public void startConnection(@NonNull InstallReferrerStateListener listener) {
    if (isReady()) {
      InstallReferrerCommons.logVerbose(
          TAG, "Service connection is valid. No need to re-initialize.");
      listener.onInstallReferrerSetupFinished(InstallReferrerResponse.OK);
      return;
    }

    if (mClientState == ClientState.CONNECTING) {
      InstallReferrerCommons.logWarn(
          TAG, "Client is already in the process of connecting to the service.");
      listener.onInstallReferrerSetupFinished(InstallReferrerResponse.DEVELOPER_ERROR);
      return;
    }

    if (mClientState == ClientState.CLOSED) {
      InstallReferrerCommons.logWarn(
          TAG, "Client was already closed and can't be reused. Please create another instance.");
      listener.onInstallReferrerSetupFinished(InstallReferrerResponse.DEVELOPER_ERROR);
      return;
    }

    // Connection to install referrer service
    InstallReferrerCommons.logVerbose(TAG, "Starting install referrer service setup.");
    mServiceConnection = new InstallReferrerServiceConnection(listener);

    Intent serviceIntent = new Intent(SERVICE_ACTION_NAME);
    serviceIntent.setComponent(new ComponentName(SERVICE_PACKAGE_NAME, SERVICE_NAME));
    List<ResolveInfo> intentServices =
        mApplicationContext.getPackageManager().queryIntentServices(serviceIntent, 0);

    if (intentServices != null && !intentServices.isEmpty()) {
      // Get component info and create ComponentName
      ResolveInfo resolveInfo = intentServices.get(0);
      if (resolveInfo.serviceInfo != null) {
        String packageName = resolveInfo.serviceInfo.packageName;
        String className = resolveInfo.serviceInfo.name;
        if (SERVICE_PACKAGE_NAME.equals(packageName)
            && className != null
            && isPlayStoreCompatible()) {
          Intent explicitServiceIntent = new Intent(serviceIntent);
          boolean connectionResult =
              mApplicationContext.bindService(
                  explicitServiceIntent, mServiceConnection, Context.BIND_AUTO_CREATE);
          if (connectionResult) {
            // Service connected successfully, listener will be called from mServiceConnection
            InstallReferrerCommons.logVerbose(TAG, "Service was bonded successfully.");
            return;
          } else {
            // Service connection was blocked, closing the connection
            // and notifying the listener
            InstallReferrerCommons.logWarn(TAG, "Connection to service is blocked.");
            mClientState = ClientState.DISCONNECTED;
            listener.onInstallReferrerSetupFinished(InstallReferrerResponse.SERVICE_UNAVAILABLE);
            return;
          }
        } else {
          // Play Store package name is not valid, ending connection
          InstallReferrerCommons.logWarn(TAG,
              "Play Store missing or incompatible. Version 8.3.73 or later required.");
          mClientState = ClientState.DISCONNECTED;
          listener.onInstallReferrerSetupFinished(InstallReferrerResponse.FEATURE_NOT_SUPPORTED);
          return;
        }
      }
    }
    // No service available to handle that Intent or service connection was blocked
    mClientState = ClientState.DISCONNECTED;
    InstallReferrerCommons.logVerbose(TAG, "Install Referrer service unavailable on device.");
    listener.onInstallReferrerSetupFinished(InstallReferrerResponse.FEATURE_NOT_SUPPORTED);
  }

  @Override
  public void endConnection() {
    mClientState = ClientState.CLOSED;
    if (mServiceConnection != null) {
      InstallReferrerCommons.logVerbose(TAG, "Unbinding from service.");
      mApplicationContext.unbindService(mServiceConnection);
      mServiceConnection = null;
    }
    mService = null;
  }

  @Override
  public ReferrerDetails getInstallReferrer() throws RemoteException {
    if (!isReady()) {
      throw new IllegalStateException(
          "Service not connected. Please start a connection before using the service.");
    }

    Bundle paramsBundle = new Bundle();
    paramsBundle.putString("package_name", mApplicationContext.getPackageName());
    try {
      ReferrerDetails referrerDetails =
          new ReferrerDetails(mService.getInstallReferrer(paramsBundle));
      return referrerDetails;
    } catch (RemoteException e) {
      InstallReferrerCommons.logWarn(TAG, "RemoteException getting install referrer information");
      mClientState = ClientState.DISCONNECTED;
      throw e;
    }
  }

  private boolean isPlayStoreCompatible() {
    PackageManager mPm = mApplicationContext.getPackageManager();
    try {
      PackageInfo info = mPm.getPackageInfo(SERVICE_PACKAGE_NAME, PackageManager.GET_META_DATA);
      int versionCode = info.versionCode;
      return versionCode >= PLAY_STORE_MIN_APP_VER;
    } catch (PackageManager.NameNotFoundException e) {
      return false;
    }
  }

  private final class InstallReferrerServiceConnection implements ServiceConnection {
    private final InstallReferrerStateListener mListener;

    private InstallReferrerServiceConnection(@NonNull InstallReferrerStateListener listener) {
      if (listener == null) {
        throw new RuntimeException("Please specify a listener to know when setup is done.");
      }
      mListener = listener;
    }

    @Override
    public void onServiceConnected(ComponentName componentName, IBinder iBinder) {
      InstallReferrerCommons.logVerbose(TAG, "Install Referrer service connected.");
      mService = IGetInstallReferrerService.Stub.asInterface(iBinder);
      mClientState = ClientState.CONNECTED;
      mListener.onInstallReferrerSetupFinished(InstallReferrerResponse.OK);
    }

    @Override
    public void onServiceDisconnected(ComponentName componentName) {
      InstallReferrerCommons.logWarn(TAG, "Install Referrer service disconnected.");
      mService = null;
      mClientState = ClientState.DISCONNECTED;
      mListener.onInstallReferrerServiceDisconnected();
    }
  }
}
